#version 120

/* DRAWBUFFERS:10 */

/*
const bool shadowHardwareFiltering0 	= false;
const bool shadowHardwareFiltering1 	= true;

const bool shadowtex0Mipmap 			= true;
const bool shadowtex1Mipmap 			= true;

const bool shadowcolor0Mipmap 			= true;
const bool shadowcolor1Mipmap 			= true;
*/

uniform vec3 upPosition;
uniform vec3 sunPosition;
uniform vec3 cameraPosition;
uniform vec3 shadowLightPosition;

uniform mat4 shadowModelView;
uniform mat4 shadowProjection;
uniform mat4 gbufferModelView;
uniform mat4 gbufferProjection;
uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferProjectionInverse;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D depthtex0;
uniform sampler2D depthtex1;
uniform sampler2D shadowtex0;
uniform sampler2D shadowcolor1;

uniform sampler2DShadow shadowtex1;
uniform sampler2DShadow shadowcolor0;

uniform float sunAngle;
uniform float frameTime;
uniform float viewWidth;
uniform float viewHeight;
uniform float aspectRatio;
uniform float rainStrength;
uniform float frameTimeCounter;

uniform int isEyeInWater;

#define UP_DIRECTION
#define LIGHT_MATRIX
#define SHADOW_MATRIX
#define FOV_PROJECTION

#include "/lib/shaderUtilities.glsl"

struct positionStruct {

    vec3 viewSpacePosition;
    vec3 worldSpacePosition;
    vec3 shadowSpacePosition;

} Position;

struct inputStruct {

    mat2x3 colorTexture2;

    vec4 colorTexture0;

    float depthTexture1;

} Input;

#include "/lib/asset/time.glsl"
#include "/lib/asset/shadow.glsl"
#include "/lib/asset/initialize.glsl"

#include "/lib/utility/misc/ditherIndexMap.glsl"

#include "/lib/utility/sky/computeSkyColor.glsl"

#include "/lib/utility/lighting/computeDiffuse.glsl"
#include "/lib/utility/lighting/computeOcclusion.glsl"
#include "/lib/utility/lighting/computeIllumination.glsl"
#include "/lib/utility/lighting/computeCrepuscularRays.glsl"

void main() {

    float curve = clamp01((FPS_TARGET * 0.75) * frameTime + 0.75) / 0.75;

    #ifndef ADAPTIVE_SCALING
        curve = 1.0;
    #endif

    vec2 compositeCoord = (texCoord / COMPOSITE_RES) * curve;
    
    Input.colorTexture2[0] = textureLoad(colortex2, texCoord).rgb;

    Input.colorTexture0 = textureLoad(colortex0, texCoord);
  
    Input.colorTexture0.r = encode2x5(abs((dither01 - 0.5) * 0.03125 + Input.colorTexture0.rb));

    if (!coordInTile(compositeCoord)) {

        pixelData0 = vec4(0.0);
        pixelData1 = vec4(Input.colorTexture0.rg, 0.0, Input.colorTexture0.a);

        return;

    } else {

        mat3x4 lightingMatrix = mat3x4(0.0);

        #define AO lightingMatrix[0].w
        #define CR lightingMatrix[1].w

        #define SD lightingMatrix[0].xyz
        #define SB lightingMatrix[1].xyz

        #define CL lightingMatrix[2]

        Input.colorTexture2[1] = textureLoad(colortex2, compositeCoord).rgb;

        ivec2 IDSource = decodeIDs(Input.colorTexture2[1].b);

        int IDs = IDSource.x;

        bool hand = bool(IDSource.y);

        Position.shadowSpacePosition = vec3(0.0);
        
        if (!hand) {

            Input.depthTexture1 = textureLoad(depthtex1, compositeCoord).r;

            Position.viewSpacePosition = projectViewSpace(compositeCoord, Input.depthTexture1, FOV, gbufferProjectionInverse);

            Position.worldSpacePosition = mat3x3(gbufferModelViewInverse) * Position.viewSpacePosition; // Cannot used optimized transMAD function due to 3D rays being offset.
            
            #if CREPUSCULAR_RAYS == 1
                computeThreeDimensionalRays(CR, IDs, compositeCoord, Position);
            #endif

            #if CREPUSCULAR_RAYS == 2
                computeTwoDimensionalRays(CR, compositeCoord);
            #endif

            if (land) {

                Position.shadowSpacePosition = distortShadowSpace(transMAD(shadowProjMatrix, transMAD(gbufferModelViewInverse, Position.viewSpacePosition)), shadowDistanceSize, inverseDistanceSize) * 0.5 + 0.5;

                vec3 normal = mat3(gbufferModelView) * decodeNormal(Input.colorTexture2[1].x, IDs);

                float skyLight = decodeHalfY(Input.colorTexture2[1].y);

                #ifdef AMBIENT_OCCLUSION
                    computeOcclusion(AO, compositeCoord, normal, skyLight, Input.depthTexture1, Position);
                #else
                    AO = 1.0;
                #endif

                #ifdef SKYLIGHT_DIFFUSE
                    computeDiffuse(SD, normal);
                #endif

                #ifdef SUNLIGHT_BOUNCE
                    if (!water) computeIllumination(SB, normal, Position);
                #endif

            }

            else

            AO = 1.0;

        }

        pixelData0 = vec4(SD * AO + SB, 1.0);
        pixelData1 = vec4(Input.colorTexture0.rg, CR, Input.colorTexture0.a);

    }

}
