#version 120

/* DRAWBUFFERS:01 */

/*
const float sunPathRotation          =  -41.7; // Sun path rotation on June 20th based in San Francisco.
const float shadowIntervalSize       =    0.5; // Every half a block in movement will update the shadow map.
const float shadowDistanceRenderMul  =    1.0;

const int noiseTextureResolution     =    1024;

const int colortex0Format = RGB10_A2;
const int colortex1Format = RGB16F;
const int colortex2Format = RGB16;
const int colortex3Format = RGBA16;

const bool colortex0Clear = false;
const bool colortex1Clear = false;
const bool colortex2Clear = false;
const bool colortex3Clear = false;

const bool colortex0MipmapEnabled   = true;
const bool colortex1MipmapEnabled   = true;
const bool colortex2MipmapEnabled   = true;

const bool shadowtex0Mipmap         = true;

const bool shadowHardwareFiltering0 = false;
const bool shadowHardwareFiltering1 = true;
*/

uniform vec3 upPosition;
uniform vec3 sunPosition;
uniform vec3 cameraPosition;
uniform vec3 shadowLightPosition;

uniform mat4 shadowModelView;
uniform mat4 gbufferModelView;
uniform mat4 shadowProjection;
uniform mat4 gbufferProjection;
uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferProjectionInverse;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D depthtex0;
uniform sampler2D depthtex1;
uniform sampler2D shadowtex0;
uniform sampler2D shadowcolor1;

uniform sampler2DShadow shadowtex1;
uniform sampler2DShadow shadowcolor0;

uniform float sunAngle;
uniform float frameTime;
uniform float viewWidth;
uniform float viewHeight;
uniform float nightVision;
uniform float rainStrength;
uniform float frameTimeCounter;

uniform int isEyeInWater;

#define UP_DIRECTION
#define LIGHT_MATRIX
#define SHADOW_MATRIX
#define FOV_PROJECTION

#include "/lib/shaderUtilities.glsl"

struct positionStruct {

    mat3x3 viewSpacePosition;
    mat3x3 worldSpacePosition;
    mat2x3 shadowSpacePosition;

} Position;

struct inputStruct {

    vec4 colorTexture0;
    vec3 colorTexture2;

    float depthTexture0;
    float depthTexture1;

} Input;

struct lightingStruct {
    
    vec3 lightShading;
    vec3 indirectLighting;
    vec3 caustic;

    mat3x3 lightmapModel;

    float shadingModel;
    float waterSurface;
    float cloudSurface;
    float rays;

} Lighting;

#include "/lib/asset/time.glsl"
#include "/lib/asset/shadow.glsl"

#include "/lib/utility/misc/ditherIndexMap.glsl"

#include "/lib/utility/sky/computeSkyColor.glsl"
#include "/lib/utility/sky/computeStars.glsl"
#include "/lib/utility/sky/computeClouds2D.glsl"
#include "/lib/utility/sky/computeClouds3D.glsl"

void main() {

    Input.colorTexture0 = textureLoad(colortex0, texCoord);
    Input.colorTexture2 = textureLoad(colortex2, texCoord).rgb;

    Input.depthTexture0 = textureLoad(depthtex0, texCoord).r;
    Input.depthTexture1 = textureLoad(depthtex1, texCoord).r;

    Position.viewSpacePosition[1] = projectViewSpace(texCoord, Input.depthTexture1, FOV, gbufferProjectionInverse);

    Position.worldSpacePosition[1] = transMAD(gbufferModelViewInverse, Position.viewSpacePosition[1]);

    ivec2 IDSource = decodeIDs(Input.colorTexture2.b);

    int IDs = IDSource.x;
    
    bool hand = bool(IDSource.y);

	vec3 output0 = vec3(0.0);
	vec3 output1 = vec3(0.0);

    Input.colorTexture0.rb = decode2x5(Input.colorTexture0.r);
   
    output0 = gammaNonLinear(Input.colorTexture0.rgb);
    
    float curve = clamp01((FPS_TARGET * 0.75) * frameTime + 0.75) / 0.75;

    #ifndef ADAPTIVE_SCALING
        curve = 1.0;
    #endif

    vec2 compositeCoord = (texCoord * COMPOSITE_RES) / curve;

    vec3 absorb = vec3(0.0);
    vec3 rayleigh = vec3(0.0);

    vec3 tone = computeTones(absorb, rayleigh);
    
    Lighting.rays = 0.0;
    Lighting.waterSurface = 0.0;
    Lighting.cloudSurface = 0.0;

	#ifdef ADVANCED_FILTERING
		float variableLOD = (1.0 - dither01) * 2;

		vec2 ditherOffset = dither01 + ivec2(1, -1);

		vec2 offsetSize = screenPixelsSize * 2;
	#else
	 	c(int) staticLOD = 2;
	#endif

    if (!hand) {   
                
        #if CREPUSCULAR_RAYS > 0
            #ifdef ADVANCED_FILTERING
                Lighting.rays = textureLoadLOD(colortex0, vec2( ditherOffset.x,  ditherOffset.x) * offsetSize + compositeCoord, variableLOD).b * 0.25 + Lighting.rays;
                Lighting.rays = textureLoadLOD(colortex0, vec2(-ditherOffset.y, -ditherOffset.y) * offsetSize + compositeCoord, variableLOD).b * 0.25 + Lighting.rays;
                Lighting.rays = textureLoadLOD(colortex0, vec2(-ditherOffset.y,  ditherOffset.x) * offsetSize + compositeCoord, variableLOD).b * 0.25 + Lighting.rays;
                Lighting.rays = textureLoadLOD(colortex0, vec2( ditherOffset.x, -ditherOffset.y) * offsetSize + compositeCoord, variableLOD).b * 0.25 + Lighting.rays;
            #else
                Lighting.rays = textureLoadLOD(colortex0, compositeCoord, staticLOD).b;
            #endif
        #endif
        
    }
    
    Position.viewSpacePosition[2] = fNormalize(Position.viewSpacePosition[1]);
    Position.worldSpacePosition[2] = fNormalize(Position.worldSpacePosition[1]);

    vec2 lightAura = vec2(0.0);

    lightAura.x = clamp01(dot(Position.viewSpacePosition[2], lightPosition[1]));
    lightAura.y = powEvenInt(lightAura.x, 4);

    if (land) {

        Position.viewSpacePosition[0] = projectViewSpace(texCoord, Input.depthTexture0, FOV, gbufferProjectionInverse);
        
        Position.worldSpacePosition[0] = transMAD(gbufferModelViewInverse, Position.viewSpacePosition[0]);

		Position.shadowSpacePosition[0] = distortShadowSpace(transMAD(shadowProjMatrix, Position.worldSpacePosition[0]), shadowDistanceSize, inverseDistanceSize) * 0.5 + 0.5;
		Position.shadowSpacePosition[1] = distortShadowSpace(transMAD(shadowProjMatrix, Position.worldSpacePosition[1]), shadowDistanceSize, inverseDistanceSize) * 0.5 + 0.5;

        Lighting.indirectLighting = vec3(0.0);

        if (hand)
        Lighting.indirectLighting = tone;

        else {
            
            #if defined SUNLIGHT_BOUNCE || defined SKYLIGHT_DIFFUSE || defined AMBIENT_OCCLUSION
                #ifdef ADVANCED_FILTERING
                    Lighting.indirectLighting = textureLoadLOD(colortex1, vec2( ditherOffset.x,  ditherOffset.x) * offsetSize + compositeCoord, variableLOD).rgb * 0.25 + Lighting.indirectLighting;
                    Lighting.indirectLighting = textureLoadLOD(colortex1, vec2(-ditherOffset.y, -ditherOffset.y) * offsetSize + compositeCoord, variableLOD).rgb * 0.25 + Lighting.indirectLighting;
                    Lighting.indirectLighting = textureLoadLOD(colortex1, vec2(-ditherOffset.y,  ditherOffset.x) * offsetSize + compositeCoord, variableLOD).rgb * 0.25 + Lighting.indirectLighting;
                    Lighting.indirectLighting = textureLoadLOD(colortex1, vec2( ditherOffset.x, -ditherOffset.y) * offsetSize + compositeCoord, variableLOD).rgb * 0.25 + Lighting.indirectLighting;
                #else
                    Lighting.indirectLighting = textureLoadLOD(colortex1, compositeCoord, staticLOD).rgb;
                #endif
            #endif

        }
        
        #if !defined SKYLIGHT_DIFFUSE
            Lighting.indirectLighting = tone * 1.1 + Lighting.indirectLighting;
        #endif
        
        bool shadowRenderZone = fLength(Position.worldSpacePosition[1]) <= shadowDistance;

		vec3 normal = mat3(gbufferModelView) * decodeNormal(Input.colorTexture2.r, IDs);

        Lighting.shadingModel = clamp01(dot(lightPosition[1], normal));

        float zDepth = sqrt(shadowPixelsSize) * 0.03125 * zScaleCoeff / (Lighting.shadingModel + 0.5);

        float zDistance = fLength(Position.viewSpacePosition[0]) / shadowDistance;

        float zCorrect = shadowPixelsSize * zDistance + zDepth;

        Position.shadowSpacePosition[0].z -= zCorrect;
        Position.shadowSpacePosition[1].z -= zCorrect;

        float absorbedLight = shadowRenderZone ? 0.25 : fDistance(Position.shadowSpacePosition[1], vec3(Position.shadowSpacePosition[1].xy, textureLoad(shadowtex0, Position.shadowSpacePosition[1].xy).r));

        if (plant || ice || glass || emissive)

        Lighting.shadingModel = absorbedLight + 0.5;

        Lighting.lightShading = vec3(0.0);

        float shadowLayers = 0.0;

        if (shadowRenderZone && !emissive) {
                
            shadowLayers = textureLoad(shadowcolor1, Position.shadowSpacePosition[1].xy).a;

            Lighting.waterSurface = shadowLoad(shadowtex1, Position.shadowSpacePosition[1]).r * Lighting.shadingModel;

                #if SHADOW_FILTER >= 1
                    vec4 blurSize = vec4(shadowPixelsSize, shadowPixelsSize, 0.0, 0.0) * (plant ? absorbedLight + 2.0 * 2.0 : 0.5);

                    sampleRCPVar(SHADOW_FILTER_SAMPLES);

                    c(float) heightSampleScale = 256.0 * sampleRCP;

                    float ditherScale = dither64 * sampleRCP;

                    #if SHADOW_FILTER == 2
                        for (int i = 0; i < SHADOW_FILTER_SAMPLES * 0.5; ++i)
                        blurSize.w = (Position.shadowSpacePosition[1].z - textureLoad(shadowtex0, hammersleyPointMap(i, sampleRCP * i + 1.0, ditherScale) * shadowPixelsSize + Position.shadowSpacePosition[1].st).r) * heightSampleScale + blurSize.w;
                        
                        blurSize.w = abs(blurSize.w) / sunAngularDiameter;

                        blurSize.xyz = blurSize.w * blurSize.xyz + blurSize.xyz;
                    #endif
                    
                    for (int i = 0; i < SHADOW_FILTER_SAMPLES; ++i)
                    Lighting.lightShading = shadowLoad(shadowtex1, vec3(hammersleyPointMap(i, sampleRCP * i + 1.0, ditherScale), 0.0) * blurSize.xyz + Position.shadowSpacePosition[1]).r * sampleRCP + Lighting.lightShading;
                #else 
                    Lighting.lightShading += shadowLoad(shadowtex1, Position.shadowSpacePosition[1]).r;
                #endif

                #ifdef SHADOW_COLORED_SHADOWS  
                    vec3 coloredShading = vec3(0.0);

                    #if SHADOW_FILTER >= 1
                        for (int i = 0; i < SHADOW_FILTER_SAMPLES; ++i) {
                            
                            vec3 currentPos = vec3(hammersleyPointMap(i, sampleRCP * i + 1.0, ditherScale), 0.0) * shadowPixelsSize + Position.shadowSpacePosition[1];

                            if (abs(textureLoad(shadowcolor1, currentPos.xy).a - 0.5) < 0.25) 

                            coloredShading = shadowLoad(shadowcolor0, currentPos).rgb * sampleRCP + coloredShading;

                        }
                    #else
                        if (abs(textureLoad(shadowcolor1, Position.shadowSpacePosition[1].xy).a - 0.5) < 0.25) 
                        
                        coloredShading = shadowLoad(shadowcolor0, Position.shadowSpacePosition[1]).rgb;
                    #endif

                    Lighting.lightShading = mix(Lighting.lightShading, coloredShading * Lighting.lightShading, fLength(coloredShading));
                #endif


        } else Lighting.lightShading = vec3(0.5);

        Lighting.indirectLighting = Lighting.indirectLighting * 0.99 + 0.01;

        Lighting.lightmapModel[0].xy = decode2x8(Input.colorTexture2.y);
        
        Lighting.lightmapModel[0].y = sqrt(Lighting.lightmapModel[0].y * mix(1.0, 0.25, clamp01(rainStrength + timeNight)));
        
        Lighting.lightmapModel[0].y = mix(Lighting.lightmapModel[0].y, 1.0, nightVision);

		Lighting.lightmapModel[1] = Lighting.lightmapModel[0].x * emissiveColor;
        
        Lighting.indirectLighting = sqrt(Lighting.indirectLighting) * Lighting.lightmapModel[0].y + (Lighting.lightmapModel[0].x * 0.25);

        Lighting.lightShading = Lighting.lightShading * Lighting.shadingModel + Lighting.indirectLighting + 0.125;
    
		Lighting.lightmapModel[2] = Lighting.lightmapModel[0].y * mix(sqrt(Lighting.indirectLighting), fLength(tone) * tone + fInverseLength(tone * mix(1.0, 4.0, clamp01(rainStrength + timeNight))), Lighting.lightShading) + 0.125;
        
        float luminance = luma(output0);

		float vibrance = RCP(luminance * exp2(saturationCoeffLn * desaturationCoeff) + saturationCoeff) * sqrt(dotX(Lighting.indirectLighting * inversesqrt(dotX(Lighting.indirectLighting * 0.75) + 0.25))) * exp2(saturationCoeff / euler) * saturationCoeff;

        output0 = saturate(output0, vibrance, luminance);

        output0 = output0 * gammaNonLinear(Lighting.lightShading * clamp01(Lighting.lightmapModel[2] + Lighting.lightmapModel[1])) + (emissive ? output0 * luma(output0) * 0.25 : vec3(0.0));

        Lighting.caustic = vec3(0.0);
        
        #ifdef SHADOW_WATER_CAUSTICS
            if (abs(shadowLayers - 0.25) < rcp8bit && shadowRenderZone) {

                if (shadowLoad(shadowtex1, Position.shadowSpacePosition[1]).r >= rcp8bit && Lighting.shadingModel == 1.0)

                output0 = shadowLoad(shadowcolor0, Position.shadowSpacePosition[1]).rgb * output0 * rayleigh;

            }
        #endif

        if (water || ice || underWater) {
        
            Lighting.waterSurface = shadowLoad(shadowtex1, Position.shadowSpacePosition[0]).r;

            if (!underWater) {

                #ifdef WATER_FOG
                    vec4 accumulateFog = vec4(0.0);

                    accumulateFog.xyz = Position.viewSpacePosition[1] - Position.viewSpacePosition[0];
                    
                    accumulateFog.w = fLength(accumulateFog.xyz) * abs(dot(normal, fNormalize(accumulateFog.xyz)));
                    
                    output0 *= exp2(-waterColor.rgb * accumulateFog.w * waterColor.a);

                    accumulateFog.w = sqrt(fExp(-accumulateFog.w / waterColor.a));

                    output0 = mix(fExp(-waterColor.rgb) * accumulateFog.w * Lighting.rays * tone * waterColor.a, output0, accumulateFog.w);
                #endif
                
            }
            
        }

    if (underWater) {
        
        #ifdef WATER_FOG
           output0 = mix(fExp(-waterColor.rgb) * Lighting.rays * tone * waterColor.a, output0, fExp(-fLength(Position.viewSpacePosition[0])));
        #endif
    }

    } else {

        vec3 skyOutput = vec3(0.0);

        vec2 horizon = vec2(0.0);

        horizon.x = powEvenInt(lightAura.y, 8) * 0.25 + clamp01(dot(Position.viewSpacePosition[2], upDirection));
        horizon.y = min(powEvenInt(horizon.x, 2), 0.25) * 2.0;

        computeSkyColor(skyOutput, Position.viewSpacePosition[2]);

        #if CLOUDS == 3 || CLOUDS == 4
            computeThreeDimensionalClouds(output1, Position.worldSpacePosition[2], rayleigh, horizon, lightAura, Lighting.cloudSurface);
        #endif

        #if CLOUDS == 2 || CLOUDS == 4
			computeTwoDimensionalClouds(output1, Position.worldSpacePosition[2], rayleigh, horizon, lightAura, Lighting.cloudSurface);
        #endif

        #ifdef STARS
            computeStars(skyOutput, Position.worldSpacePosition[2], horizon, Lighting.cloudSurface);
        #endif

        output0 = mix(skyOutput, output0 * fLength(tone + rcp8bit), 1.0 - Input.colorTexture0.a);

    }

    output0 = gammaNonLinear(Lighting.rays * rayleigh * mix(lightAura.y * 2.0 + 0.5, lightAura.y * 0.25 + 0.25, timeNight)) + output0;

    output0 = gammaLinear(output0);

    pixelData0 = vec4(output0, land ? Lighting.waterSurface : Lighting.cloudSurface);
    pixelData1 = vec4(clamp01(output1), 1.0);
}