#version 120

/* DRAWBUFFERS:12 */

/*
const bool colortex0MipmapEnabled = true;
const bool colortex1MipmapEnabled = true;
*/

uniform vec3 upPosition;
uniform vec3 sunPosition;
uniform vec3 cameraPosition;
uniform vec3 shadowLightPosition;

uniform sampler2D noisetex;
uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D depthtex0;
uniform sampler2D depthtex1;

uniform mat4 gbufferModelView;
uniform mat4 gbufferProjection;
uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferProjectionInverse;

uniform float sunAngle;
uniform float viewWidth;
uniform float viewHeight;
uniform float rainStrength;
uniform float frameTimeCounter;

uniform int isEyeInWater;

#define UP_DIRECTION
#define LIGHT_MATRIX
#define FOV_PROJECTION

#include "/lib/shaderUtilities.glsl"

mat4x3 reflectionMatrix = mat4x3(0.0);

struct inputStruct {

    vec3 colorTexture0;
    vec3 colorTexture1;
    vec3 colorTexture2;

    float depthTexture0;
    float depthTexture1;

} Input;

struct positionStruct {

	mat3x3 viewSpacePosition;

	vec3 worldSpacePosition;

	mat2x3 reflectedRay;

} Position;


#include "/lib/asset/time.glsl"

#include "/lib/utility/misc/ditherIndexMap.glsl"

#include "/lib/utility/sky/computeSkyColor.glsl"
#include "/lib/utility/sky/computeStars.glsl"
#include "/lib/utility/sky/computeClouds2D.glsl"
#include "/lib/utility/sky/computeClouds3D.glsl"

#include "/lib/utility/water/waterWaves.glsl"
#include "/lib/utility/water/waterParallax.glsl"
#include "/lib/utility/water/computeRaytracedCoordinate.glsl"

#include "/lib/utility/refract/computeRefractionComposite2.glsl"

void main() {

    Input.colorTexture0 = textureLoad(colortex0, texCoord).rgb;
    Input.colorTexture1 = textureLoad(colortex1, texCoord).rgb;
    Input.colorTexture2 = textureLoad(colortex2, texCoord).rgb;

    Input.depthTexture0 = textureLoad(depthtex0, texCoord).r;

    Position.viewSpacePosition[0] = projectViewSpace(texCoord, Input.depthTexture0, FOV, gbufferProjectionInverse);

    Position.viewSpacePosition[2] = fNormalize(Position.viewSpacePosition[0]);

    Position.reflectedRay = mat2x3(0.0);

    ivec2 IDSource = decodeIDs(Input.colorTexture2.b);

    int IDs = IDSource.x;

    vec3 output0 = gammaNonLinear(Input.colorTexture0);
    vec3 output1 = gammaNonLinear(Input.colorTexture1);
    
    float sky = 0.0;
    float reflectedIDs = 0.0;

    vec3 absorb = vec3(0.0);
    vec3 rayleigh = vec3(0.0);

    vec3 tone = computeTones(absorb, rayleigh);
    
	if (underWater) {

        #ifdef REFRACTION
            #ifdef REFRACTION_UNDERWATER_DISTORTION
                computeUnderwaterRefraction(output0, Input.colorTexture2);
            #endif
        #endif

    }

    if (!land) {

        output0 = textureLoadLOD(colortex1, texCoord, 1 + dither01).rgb + output0;

        sky = smoothConstStep(sunAngularDiameterCos , 2974.25075129, dot(Position.viewSpacePosition[2],  lightPosition[0])) * sunLuminance  + sky;
        sky = smoothConstStep(moonAngularDiameterCos, 2997.91456069, dot(Position.viewSpacePosition[2], -lightPosition[0])) * moonLuminance + sky;
             
        sky = sky * (1.0 - textureLoad(colortex0, texCoord).a);

        output0 = sky * absorb + output0;

    }

    if (water) {

        Input.depthTexture1 = textureLoad(depthtex1, texCoord).r;

        Position.viewSpacePosition[1] = projectViewSpace(texCoord, Input.depthTexture1, 1.0, gbufferProjectionInverse);

        Position.worldSpacePosition = transMAD(gbufferModelViewInverse, Position.viewSpacePosition[0]);

        vec3 normal = mat3(gbufferModelView) * decodeNormal(Input.colorTexture2.r, IDs);

        mat2x3 waves = mat2x3(normal, normal);

        #if defined WATER_NORMALS || defined REFRACTION_WATER_WAVES
            vec3 tangent = cross(normal, gbufferModelView[2].xyz);
            
            mat3 tbnMatrix = transpose(mat3(tangent, cross(tangent, normal), normal));

            mat2x3 position = mat2x3(0.0);

            position[0] = Position.worldSpacePosition + cameraPosition;
            position[1] = tbnMatrix * Position.viewSpacePosition[2];

            #ifdef WATER_PARALLAX_WAVES
                getParallaxWaves(position);
            #endif

            waves[0].y = getWaves(position[0]);
            waves[0].x = getWaves(position[0] + vec3(wavesDifferentialCoeff, 0.0, 0.0));
            waves[0].z = getWaves(position[0] + vec3(0.0, 0.0, wavesDifferentialCoeff));

            waves[1].xy = (waves[0].y - waves[0].xz) / wavesDifferentialCoeff;

            waves[1].z = sqrt(1.0 - dotX(waves[1].xy));

            waves[0] = waves[1];
            waves[1] = waves[1] * tbnMatrix;
        #endif

        #ifdef REFRACTION 
            #ifdef REFRACTION_WATER_WAVES
                computeWaterRefraction(output0, waves, Position);
            #endif
        #endif
        
        #ifdef WATER_REFLECTIONS
            #define rayCoord reflectionMatrix[0].rgb

            #define skyOutput reflectionMatrix[1].rgb
            #define reflectedOutput0 reflectionMatrix[2].rgb
            #define reflectedOutput1 reflectionMatrix[3].rgb

            #define reflection reflectionMatrix[2].a

            float skyLight = decode2x8(Input.colorTexture2.y).y;

            Position.viewSpacePosition[1] = Position.viewSpacePosition[0];

            Position.reflectedRay[0] = reflect(Position.viewSpacePosition[2], waves[1]);
            Position.reflectedRay[1] = mat3(gbufferModelViewInverse) * Position.reflectedRay[0];

            float fresnel = powEvenInt(clamp01(1.0 + dot(waves[1], Position.viewSpacePosition[2])), 4) * 0.9 + 0.1;

            float cloudSurface = 0.0;

            int refinements = 0;

            bool hit = false;

            computeRaytracedCoordinates(rayCoord, hit, refinements, Position);

            if (hit) {

                reflectedIDs = decodeIDs(textureLoad(colortex2, rayCoord.xy).b).x;

                #if WATER_REFLECTION_TYPE == 0
                    reflectedOutput0 = gammaNonLinear(textureLoad(colortex0, rayCoord.xy).rgb);
                #else
                    sampleRCPVar(WATER_ANISOTROPIC_SAMPLES);

                    vec2 stretchSize = sampleRCP * fresnel * screenPixelsSize * ivec2(4, 32);
                    
                    c(int) halfSamples = WATER_ANISOTROPIC_SAMPLES / 2;

                    for (int i = 0; i < WATER_ANISOTROPIC_SAMPLES; ++i)
                    reflectedOutput0 = gammaNonLinear(textureLoad(colortex0, rotatePoint(i - halfSamples) * stretchSize + rayCoord.xy).rgb) * sampleRCP + reflectedOutput0;
                #endif

            }
            
            if (skyLight > 0.5 && reflectedIDs == 0) {
                
                vec2 lightAura = vec2(0.0);
            
                lightAura.x = clamp01(dot(Position.reflectedRay[0], lightPosition[1]));
                lightAura.y = powEvenInt(lightAura.x, 4);
                
                vec2 horizon = vec2(0.0);
            
                horizon.x = powEvenInt(lightAura.y, 8) * 0.25 + clamp01(dot(Position.reflectedRay[0], upDirection));
                horizon.y = min(powEvenInt(horizon.x, 2), 0.25);

                computeSkyColor(skyOutput, Position.reflectedRay[0]);

                #if CLOUDS == 3 || CLOUDS == 4
                    computeThreeDimensionalClouds(reflectedOutput1, Position.reflectedRay[1], rayleigh, horizon, lightAura, cloudSurface);
                #endif

                #if CLOUDS == 2 || CLOUDS == 4
                    computeTwoDimensionalClouds(reflectedOutput1, Position.reflectedRay[1], rayleigh, horizon, lightAura, cloudSurface);
                #endif

                #ifdef STARS
                    computeStars(reflectedOutput1, Position.reflectedRay[1], horizon, cloudSurface);
                #endif

                #if CREPUSCULAR_RAYS > 0
                    skyOutput = gammaNonLinear(0.5 * rayleigh * mix(lightAura.y * 2.0 + 0.5, lightAura.y * 0.25 + 0.25, timeNight)) + skyOutput;
                #endif

                reflectedOutput0 = skyOutput + reflectedOutput1;

            }

            if (textureLoad(colortex0, texCoord).a > 0.0 && reflectedIDs == 0) {

                sky = smoothConstStep(sunAngularDiameterCos , 2974.25075129, dot(Position.reflectedRay[0],  lightPosition[0])) * sunLuminance  + sky;
                sky = smoothConstStep(moonAngularDiameterCos, 2997.91456069, dot(Position.reflectedRay[0], -lightPosition[0])) * moonLuminance + sky;

                sky = sky * (1.0 - cloudSurface);

            }

            output0 = mix(sky * absorb + output0, reflectedOutput0, fresnel);
        #endif

    }

    if (emissive || emissiveCheck(reflectedIDs)) output0 = output0 + (sunLuminance * 0.04 * luma(output0)) * gammaNonLinear(output0);

    output0 = gammaLinear(output0);

    pixelData0 = vec4(output0, 1.0);
    pixelData1 = vec4(warm, decode2x4(Input.colorTexture2.b).y > 0, textureLoad(noisetex, texCoord).r, 1.0);
}