#version 120

/* DRAWBUFFERS:01 */

/*
const bool colortex1MipmapEnabled = true;
const bool colortex2MipmapEnabled = true;
*/

uniform sampler2D colortex1;
uniform sampler2D colortex2;

uniform float viewWidth;
uniform float viewHeight;
uniform float aspectRatio;
uniform float rainStrength;
uniform float frameTimeCounter;

uniform ivec2 eyeBrightness;

#define LENS_SETTINGS
#define BLOOM_TILE_COMPUTE

#include "/lib/shaderUtilities.glsl"

struct inputStruct {

    vec3 colorTexture1;

} Input;

#include "/lib/utility/misc/ditherIndexMap.glsl"

#include "/lib/utility/post/computeLensSimulations.glsl"
#include "/lib/utility/post/computeBloomTile.glsl"

#include "/lib/utility/refract/computeRefractionComposite3.glsl"

void main() {

    float LOD = 0.0;

    mat3x2 newCoord = mat3x2(texCoord, texCoord, texCoord);
    
    #define coord newCoord[0]

    #ifdef REFRACTION
        computeRefraction(newCoord, LOD);
    #endif

    Input.colorTexture1.r = textureLoadLOD(colortex1, newCoord[0], LOD).r;
    Input.colorTexture1.g = textureLoadLOD(colortex1, newCoord[1], LOD).g;
    Input.colorTexture1.b = textureLoadLOD(colortex1, newCoord[2], LOD).b;
    
    vec3 output0 = vec3(0.0);

    vec2 tileDistance = screenPixelsSize * 256.0 + 2.0;

    #ifdef BLOOM
    	computeBloomTile(output0, 0);
		computeBloomTile(output0, 1);
		computeBloomTile(output0, 2);

		if (BLOOM_TILES >= 4) computeBloomTile(output0, 3);
		if (BLOOM_TILES >= 5) computeBloomTile(output0, 4);
		if (BLOOM_TILES >= 6) computeBloomTile(output0, 5);
		if (BLOOM_TILES >= 7) computeBloomTile(output0, 6);
    #endif

    #ifdef LENS_GLARE
        coord = downScale(texCoord, vec2(2.0, 2.0), vec2(1.0, 1.0));

        if (coordInTile(coord)) {

            c(vec2) coordOffset = LENS_GLARE_TYPE > 0 ? vec2(1.0, 0.0) : vec2(0.0, 1.0);
            
            vec2 indexSize = dither64 * screenPixelsSize;

            output0 *= 2.0;

            output0 = gammaNonLinear(textureLoadLOD(colortex1,  coordOffset * indexSize + coord, 0).rgb) + output0;
            output0 = gammaNonLinear(textureLoadLOD(colortex1, -coordOffset * indexSize + coord, 0).rgb) + output0;

            output0 *= 0.5;

        }
    #endif

    #if LENS_FLARE_TYPE > 1 || defined LENS_DIFFRACTION
        coord = downScale(texCoord, tileDistance, vec2(tileDistance.x - 1.0, 0.0));

        if (coordInTile(coord)) {
            
            #if LENS_FLARE_TYPE > 1
                float weight = fPow(1.0 - fLength(abs(coord - 0.5)), 16.0);

                if (weight > rcp8bit) {

                    vec2 offset = coord - 0.5;

                    computeFlare(output0, vec2(1.9, 1.95), vec2(1.1, 1.1), coord, offset);

                    computeFlare(output0, vec2(1.4, 1.45), vec2(1.3, 1.3), coord, offset);

                    computeFlare(output0, vec2(0.7, 0.72), vec2(1.3, 1.3), coord, offset);

                    output0 *= weight;
                        
                }
            #endif
            
            #ifdef LENS_DIFFRACTION
                computeDiffraction(output0, coord);
            #endif

        }
    #endif

    #ifdef LENS_DIRT
        coord = downScale(texCoord, vec2(tileDistance.x, 2.0), vec2(0.0, 1.0));

        if (coordInTile(coord)) computeDirtyLens(output0, coord);
    #endif

    output0 = gammaLinear(output0);

    c(float) overflowRange = 4.59479341999; // 2 ^ 2.2 = x
    c(float) overflowBase = 0.683020128; // overflowRange ^ -0.25 = x
    c(float) overflow = 0.66226599046; // e ^ x = overflowRange
    c(float) lift = 0.02670782484; // 2.2 ^ overflowRange

    float luminance = luma(output0);

    output0 = output0 * (luminance * fExp(-(fPow(luminance, overflow) * overflow + overflowBase)) + lift);

    pixelData0 = vec4(Input.colorTexture1, textureLoad(colortex2, texCoord).g);
    pixelData1 = vec4(output0, 1.0);
}