#version 120

uniform vec3 cameraPosition;
uniform vec3 previousCameraPosition;

uniform mat4 gbufferProjection;
uniform mat4 gbufferModelViewInverse;
uniform mat4 gbufferProjectionInverse;
uniform mat4 gbufferPreviousModelView;
uniform mat4 gbufferPreviousProjection;

uniform sampler2D colortex1;
uniform sampler2D depthtex1;

uniform float frameTime;
uniform float viewWidth;
uniform float viewHeight;
uniform float frameTimeCounter;

uniform int isEyeInWater;

#include "/lib/shaderUtilities.glsl"

struct inputStruct {

    vec3 colorTexture1;

} Input;

#include "/lib/utility/misc/ditherIndexMap.glsl"
#include "/lib/utility/misc/computeCameraVelocity.glsl"

#include "/lib/utility/post/computeMotionblur.glsl"
#include "/lib/utility/post/computeTonemapping.glsl"

void main() {
    
    Input.colorTexture1 = textureLoad(colortex1, texCoord).rgb;

    #if defined MOTIONBLUR && ANTI_ALIASING <= 1
        if (textureLoad(colortex1, texCoord).a > 0) computeMotionBlur(Input.colorTexture1);
    #endif

    Input.colorTexture1 = gammaLinear(ACESFilm(gammaNonLinear(Input.colorTexture1) * 2.0));

    Input.colorTexture1 *= blackbody(COLOR_TEMPERATURE);

    if (COLOR_FILTER > 0) {

        float luminance = luma(Input.colorTexture1);

        if (COLOR_FILTER == 1) Input.colorTexture1 = lumLerp(Input.colorTexture1, luminance, vec3(1.00, 0.85, 0.75));
        if (COLOR_FILTER == 2) Input.colorTexture1 = lumLerp(Input.colorTexture1, luminance, vec3(0.75, 0.85, 1.00));
        if (COLOR_FILTER == 3) Input.colorTexture1 = lumLerp(Input.colorTexture1, luminance, vec3(0.65, 0.30, 0.10));
        if (COLOR_FILTER == 4) Input.colorTexture1 = lumLerp(Input.colorTexture1, luminance, vec3(1.00, 1.00, 1.00));

    }

    pixelColor = vec4(Input.colorTexture1, 1.0);
}