void computeThreeDimensionalRays(io float CR3D, in int IDs, in vec2 compositeCoord, in positionStruct Position) {

    Position.worldSpacePosition = fNormalize(Position.worldSpacePosition);

    Position.shadowSpacePosition = transMAD(shadowProjMatrix, Position.worldSpacePosition + gbufferModelViewInverse[3].xyz);
    
    vec3 shadowStep = diagonal3(shadowProjection) * (mat3(shadowModelView) * Position.worldSpacePosition);

    float playerDistance = fLength(Position.viewSpacePosition);

    c(float) rayIncrement = 64.0 / CREPUSCULAR_RAYS_SAMPLES;

    #if SHADOW_FILTER >= 1 && defined CREPUSCULAR_RAYS_SHADOW_FILTER
        c(float) pixelBlurSize = shadowPixelsSize * SHADOW_FILTER_BLUR_SIZE;

        vec4 blurSize = vec4(pixelBlurSize, pixelBlurSize, 0.0, 0.0);

        c(int) sampleCount = SHADOW_FILTER_SAMPLES / 4;

        sampleRCPVar(sampleCount);

        c(float) heightSampleScale = 256.0 * sampleRCP;

        float ditherScale = dither64 * sampleRCP;
    #endif 

    for (float i = dither01 * rayIncrement + RCP(rayIncrement); i < CREPUSCULAR_RAYS_3D_DEPTH && playerDistance > i; i += rayIncrement) {

        vec3 samplePosition = distortShadowSpace(shadowStep * i + Position.shadowSpacePosition, shadowDistanceSize, inverseDistanceSize) * 0.5 + 0.5;

        #if SHADOW_FILTER >= 1 && defined CREPUSCULAR_RAYS_SHADOW_FILTER
            #if SHADOW_FILTER == 2
                for (int j = 0; j < sampleCount; ++j)
                blurSize.w = (samplePosition.z - textureLoad(shadowtex0, hammersleyPointMap(j, sampleRCP * j + 1.0, ditherScale) * pixelBlurSize + samplePosition.st).r) * heightSampleScale + blurSize.w;
                
                blurSize.w = abs(blurSize.w) / sunAngularDiameter;

                blurSize.xyz = blurSize.w * blurSize.xyz + blurSize.xyz;
            #endif
            
            for (int j = 0; j < sampleCount; ++j)
            CR3D = shadowLoad(shadowtex1, vec3(hammersleyPointMap(j, sampleRCP * j + 1.0, ditherScale), 0.0) * blurSize.xyz + samplePosition).r * sampleRCP + CR3D;

            blurSize = vec4(pixelBlurSize, pixelBlurSize, 0.0, 0.0);
        #else
            CR3D += shadowLoad(shadowtex1, samplePosition).r;
        #endif

        #if defined SHADOW_WATER_CAUSTICS && defined CREPUSCULAR_RAYS_CAUSTIC_RAYS
            if (abs(textureLoad(shadowcolor1, samplePosition.xy).a - 0.25) < rcp8bit)
            CR3D = shadowLoad(shadowcolor0, samplePosition).r * clamp01(samplePosition.z - textureLoad(shadowtex0, samplePosition.xy).r) * CR3D + CR3D;
        #endif

    }

    CR3D = RCP(CREPUSCULAR_RAYS_3D_DEPTH * 2.0) * rayIncrement * CR3D;

}

void computeTwoDimensionalRays(io float CR2D, in vec2 coord) {

    c(float) sampleRCP = RCP(CREPUSCULAR_RAYS_SAMPLES);

	vec3 projectedLightPosition = projMAD3(gbufferProjection, lightPosition[0]);

    vec2 deltaPosition = (coord - ((projectedLightPosition.xy / projectedLightPosition.z) * 0.5 + 0.5)) * sampleRCP;

	for (int i = 0; i < CREPUSCULAR_RAYS_SAMPLES; ++i)
    CR2D = step(decodeHalfX(textureLoad(colortex2, deltaPosition * dither01 + (coord - deltaPosition * i)).b), 0.0) * sampleRCP + CR2D;

    CR2D *= 0.5;
} 