void computeOcclusion(io float AO, in vec2 coord, in vec3 normal, in float skyLight, in float depthInput, in positionStruct Position) {

    depthInput = 0.1 / (1.1 - depthInput);

    sampleRCPVar(AMBIENT_OCCLUSION_SAMPLES);

    float ditherScale = dither64 * sampleRCP;

    vec2 projectionRadius = 0.125 * inversesqrt(dotX(Position.viewSpacePosition) * 0.5 + 1.0) + screenPixelsSize;
      
    vec3 sunDirection = lightPosition[1] * skyLight;
    vec3 skyDirection = upDirection * skyLight;
    
    vec2 offsetCoord = vec2(0.0);

    vec3 directionOffset = vec3(0.0);
    
    float occlusion = 0.0;

    for (int i = 0; i < AMBIENT_OCCLUSION_SAMPLES; ++i) {

        vec2 offsetMap = hammersleyPointMap(i, sampleRCP * i + 1.0, ditherScale);

        #ifdef AMBIENT_OCCLUSION_DIRECTIONAL
            offsetCoord = offsetMap * projectionRadius + Position.shadowSpacePosition.xy;

            directionOffset = sunDirection * textureLoad(shadowtex0, offsetCoord).r + skyDirection;
        #endif

        offsetCoord = (directionOffset.xy * (i * sampleRCP) + offsetMap) * projectionRadius + coord;

        float offsetDepth = textureLoad(depthtex1, offsetCoord).r;

        vec3 offsetViewSpacePosition = projectViewSpace(offsetCoord, offsetDepth, FOV, gbufferProjectionInverse);

        occlusion = abs(dot(normal, fNormalize(offsetViewSpacePosition.xyz - Position.viewSpacePosition.xyz)) - 1.0);
        
        occlusion =  max(occlusion, abs((0.1 / (1.1 - offsetDepth)) - depthInput) * fDistanceRCP(coord, offsetCoord));
    
        AO = occlusion * sampleRCP + AO;
    }

    AO = clamp01(AO);
    
}