#version 120

/* DRAWBUFFERS:01 */

#define PIXEL

varying vec4 colorInput;

varying vec3 vertexAttributes;

varying vec2 lightmapInput;

uniform vec3 cameraPosition;

uniform sampler2D texture;

uniform float frameTimeCounter;

#include "/lib/shaderUtilities.glsl"

#include "/lib/utility/water/waterWaves.glsl"

flat(vec3) normalInput;

flat(float) IDInput;

void main() {

    vec4 output0 = texture2D(texture, texCoord) * colorInput;

    float IDs = round(IDInput);

    float layer = 0.0;
   
    if (glass) {

        #if GLASS_LAYERS == 0
            output0.a = 1.0;
        #elif GLASS_LAYERS == 1
            output0.a = floor(output0.a + 0.5) - 0.5;
        #endif

    }
    
    if (colored)

    layer = 0.5;

    #ifdef SHADOW_WATER_CAUSTICS
        if (waterSource) {

            mat2x3 waves = mat2x3(0.0);
            
            waves[0].y = getWaves(vertexAttributes);
            waves[0].x = getWaves(vertexAttributes + vec3(wavesDifferentialCoeff, 0.0, 0.0));
            waves[0].z = getWaves(vertexAttributes + vec3(0.0, 0.0, wavesDifferentialCoeff));

            waves[1].xy = (waves[0].y - waves[0].xz) / wavesDifferentialCoeff;

            waves[1].z = sqrt(1.0 - dotX(waves[1].xy));

            waves[1].z = dot(waves[1].xy, waves[1].zz);

            waves[1] = fNormalize(waves[1]);

            vec3 normalizedPos = fNormalize(vertexAttributes);

            vec3 refractPos = refract(normalizedPos, waves[1], 0.75);
            
            float caustic = fPow(fLength(dFdx(normalizedPos) * dFdy(normalizedPos)) * fInverseLength(dFdx(refractPos) * dFdy(refractPos)), 0.125);

            output0.rgb = vec3(caustic);
            
            layer = 0.25;

        }
    #endif

    pixelData0 = output0;
    pixelData1 = vec4(normalInput.xy * 0.5 + 0.5, lightmapInput.y, layer);
}